# coding:utf-8
"""Views Module."""
import inspect
import time

from tarpit.config import CONFIG as O_O
from tarpit.web import BaseController, route, auth
from tarpit.utils.totp import get_totp_token
from tarpit.checker import P, M


@route(r'/')
class Account(BaseController):
    """account"""

    @auth
    async def get(self, *_args, **_kw):

        result = await self.miss.user.get_account_info(
            user_id=self.token_args.user_id)

        self.success(result)

    @auth
    async def put(self, *_args, **_kw):

        args = self.check_params({
            P('name') >> str,
            P('phone') >> str,
            P('avatar') >> str,
        })

        result = await self.miss.user.modify_account_info(
            user_id=self.token_args.user_id,
            name=args.name,
            phone=args.phone,
            avatar=args.avatar,
        )

        self.success(result)


@route(r'/modify_password')
class AccountModifyPassword(BaseController):
    """acccount modify password"""

    @auth
    async def post(self, *_args, **_kw):

        args = self.check_params({
            M('user_id') >> int,
            M('oldpass') >> str,
            M('newpass') >> str,
        })

        user_info = await self.miss.user.check_password_by_id(
            user_id=self.token_args.user_id, password=args.oldpass)

        if not user_info:
            self.fail(401)

        result = await self.miss.set_password(
            user_id=args.user_id,
            password=args.password,
        )

        self.success(result)


@route(r'/address')
class AccountAddress(BaseController):
    """account address"""

    @auth
    async def get(self, *_args, **_kw):

        args = self.check_params({
            M('address_id') >> int,
        })

        result = await self.miss.address.get_address_info(
            address_id=args.address_id)

        self.success(result)

    @auth
    async def post(self, *_args, **_kw):

        args = self.check_params({
            M('recipient') >> str,
            M('phone') >> str,
            M('address') >> str,
        })

        await self.miss.address.create_address(
            user_id=self.token_args.user_id,
            recipient=args.recipient,
            phone=args.phone,
            address=args.address)

        self.success()

    @auth
    async def put(self, *_args, **_kw):

        args = self.check_params({
            M('address_id') >> int,
            P('recipient') >> str,
            P('phone') >> str,
            P('address') >> str,
        })

        await self.miss.address.modify_address_info(
            address_id=args.address_id,
            recipient=args.recipient,
            phone=args.phone,
            address=args.address)

        self.success()

    @auth
    async def delete(self, *_args, **_kw):

        args = self.check_params({
            M('address_id') >> int,
        })

        await self.miss.address.delete_address(address_id=args.address_id)

        self.success()


@route(r'/address/list')
class AccountAddressList(BaseController):
    """account list"""

    @auth
    async def get(self, *_args, **_kw):

        result = await self.miss.address.get_address_list(
            user_id=self.token_args.user_id)

        self.success(dict(infos=result))


@route(r'/signin')
class SignIn(BaseController):
    """sign in"""

    async def post(self, *_args, **_kw):

        args = self.check_params({
            M('username') >> str,
            M('password') >> str,
        })

        user_info = await self.miss.user.check_password(
            username=args.username, password=args.password)

        if not user_info:
            self.fail(404)
        if user_info.role > 0:
            self.fail(401)

        self.set_token(user_info)
        self.set_header('Access-Control-Expose-Headers',
                        O_O.server.token_header)

        self.success(user_info)



from .checker import P, M
from .web import route, auth, BaseController
from .task import service
from .dao import MongoBase
from .config import CONFIG as conf

__all__ = [
    'P',
    'M',
    'route',
    'auth',
    'service',
    'conf',
    'BaseController',
    'MongoBase',
]



# coding:utf-8

import json
from json import JSONEncoder


def tarpit_default(o):
    try:
        return iter(o)
    except TypeError:
        return {k: str(o.__dict__[k]) for k in o.__dict__}


json._default_encoder = JSONEncoder(
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    indent=None,
    separators=None,
    default=tarpit_default,
)


def tarpit_dumps(obj):
    return json.dumps(obj, default=tarpit_default)


def tarpit_loads(obj):
    return json.loads(obj)



# coding:utf-8
"""Views Module."""
import inspect
import time

from tarpit.config import CONFIG as O_O
from tarpit.toolkit import P, route, BaseController


@route(r'/')
class Index(BaseController):
    """Test index request handler."""

    async def get(self, *_args, **_kwargs):
        """Get method of IndexHandler."""

        args = self.check_params({
            P('a') >> str,
            P('b') >> int,
        })

        self.finish(f'<h1>Dejavu! {args.a} {args.b}</h1>'.encode())


@route(r'*.php')
class PHPScan(BaseController):
    """Test index request handler."""

    async def get(self, *_, **__):
        """return f word."""

        self.finish(b'fuck you')



import os
import sys
import shutil
import pkgutil
import importlib
from tornado import ioloop
from tornado.web import Application
from tornado.log import enable_pretty_logging

from .web import ROUTES
from .config import CONFIG as O_O


def recursive_import(submodule_location, submodule_name=''):
    dirname = submodule_name and (submodule_name + '.')
    module_dict = dict()
    for m in pkgutil.iter_modules(submodule_location):
        if m.ispkg:
            location = f'{submodule_location[0]}/{m.name}'
            recursive_import([location], submodule_name=f'{dirname}{m.name}')
        else:
            module_dict[m.name] = importlib.import_module(f'{dirname}{m.name}')


def set_dependence(module):
    recursive_import(module.__spec__.submodule_search_locations,
                     module.__name__)


def generate(dst):
    src = os.path.join(os.path.dirname(__file__), 'templates')
    shutil.copytree(src, dst)


def serve():

    sys.path.append(os.getcwd())
    mappers = __import__('mappers')
    services = __import__('services')
    controllers = __import__('controllers')
    storage = __import__('storage')
    set_dependence(storage)
    set_dependence(mappers)
    set_dependence(services)
    set_dependence(controllers)

    enable_pretty_logging()

    for r in ROUTES:
        print(f'{r[0]:30s} {r[1]}')

    app = Application(handlers=ROUTES, **O_O.application)
    app.listen(O_O.server.port, **O_O.httpserver)

    O_O.show()
    print('\nstart listen...')
    ioloop.IOLoop.instance().start()


# coding:utf-8
"""Lazor Database Module."""
import uuid
import enum
from datetime import datetime
from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (CHAR, BigInteger, Column, Enum, Integer, SmallInteger,
                        Text, String, Numeric, Float, TIMESTAMP)
from sqlalchemy import PrimaryKeyConstraint, Sequence, UniqueConstraint
from sqlalchemy import func, text

Base = declarative_base()


class CreatedAt:
    created_at = Column(TIMESTAMP, server_default=text('CURRENT_TIMESTAMP'))


class UpdatedAt:
    updated_at = Column(
        TIMESTAMP,
        server_default=text('CURRENT_TIMESTAMP  ON UPDATE CURRENT_TIMESTAMP'))


class Remark:
    remark = Column(Text)


class Deleted:
    deleted = Column(SmallInteger, server_default=text('0'))



from tarpit.dao import MongoBase


class MongoSession(MongoBase):
    """Mongo Template."""

    __collection_name__ = 'session'
    __alias__ = 'session'  # optional

    def __collection_init__(self):
        """Write some collection initial operation, like set some index."""
        self.sync_client.storage.drop_indexes()
        self.sync_client.storage.create_index(
            'timestamp', expireAfterSeconds=86400)
        self.sync_client.storage.create_index('mail')

    async def insert_image(self):
        await self.client.image.update_one(
            {'session_id': 1}, {'$set': {
                'user_id': 2,
                'user_name': 'lkjklj'
            }},
            upsert=True)



# coding:utf-8
"""Base module for other views' modules."""
import json
import re
import time
import traceback
import os
import pkgutil
import importlib
import asyncio
from functools import wraps
from collections import abc
from urllib import parse
import inspect

import jwt
from tornado import gen, httpclient
from tornado.web import Finish, RequestHandler
from tornado.log import app_log, gen_log

from .err import *
from .config import CONFIG as O_O, get_status_message
from .utils import Arguments, dump_in, dump_out, dump_error, json_encoder
from .task import TaskError, SERVICES
from .dao import MongoFactory

ROUTES = []


def route(*path_list: str):
    """Decorator to bind path to a hander."""

    def wrapper(handler: RequestHandler):
        if not issubclass(handler, RequestHandler):
            raise PermissionError('Can\'t routing a nonhandler class.')

        filename = traceback.extract_stack(limit=2)[0].filename
        if '/controllers' in filename:
            filedir = os.path.dirname(filename).split('/controllers')[-1]
            filedir = filedir.strip('/')
            filename = os.path.basename(filename)
            filename = os.path.splitext(filename)[0]
            filename = '' if filename == 'index' else filename
            for path in path_list:
                realpath = path.strip('/')
                realpath = '/' + f'{filedir}/{filename}/{realpath}'.strip('/')
                ROUTES.append((realpath, handler))
        return handler

    return wrapper


class auth:

    def __init__(self, optional=False, **options):
        self.options = options
        self.optional = optional

    def __new__(cls, func=None, **options):
        instance = super(auth, cls).__new__(cls)
        instance.__init__(**options)
        if func:
            return instance(func)
        else:
            return instance

    def __call__(self, func):
        """Decorator to check authorization of user."""

        def process(ctlr):
            ctlr.set_header('Access-Control-Allow-Origin', '*')
            token_params = Arguments(ctlr.get_token())
            now = int(time.time())
            if not self.optional:
                if not token_params:
                    raise AuthenticationError('Token is empty.')
                if 'timestamp' not in token_params:
                    raise AuthenticationError('No timestamp info in token.')
                if token_params.timestamp < now:
                    raise TokenExpiredError()

                for key in self.options:
                    if isinstance(self.options[key], abc.Iterable):
                        if token_params.get(key) not in self.options[key]:
                            raise AuthorizationError(key)
                    else:
                        if token_params.get(key) != self.options[key]:
                            raise AuthorizationError(key)

            ctlr.token_args = Arguments(token_params)

        @wraps(func)
        async def async_wrapper(ctlr, **kwargs):
            process(ctlr)
            await func(ctlr, **kwargs)

        @wraps(func)
        def wrapper(ctlr, **kwargs):
            process(ctlr)
            func(ctlr, **kwargs)

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        else:
            return wrapper


class Mission:

    def __init__(self, func, controller):
        if not callable(func):
            self.func = self.none_func
            self.func_name = 'none_func'
            raise TypeError(f'Need callable Object, not {type(func)}')
        self.func_name = func.__name__
        self.func = func
        self.controller = controller

    @staticmethod
    def none_func(*args, **kwargs):
        return 'Uncallable function.'

    async def __call__(self, *args, **kwargs):

        if not O_O.worker.mode:
            return self.run(*args, **kwargs)

        async_task = self.func.apply_async(args=args, kwargs=kwargs)

        while not async_task.ready():
            asyncio.sleep(0)

        if async_task.failed():
            dump_error(f'Task Failed: {self.func_name}[{async_task.task_id}]',
                       f'    {str(async_task.result)}')
            raise RuntimeError(str(async_task.result))
        else:
            result = async_task.result

        if isinstance(result, dict) and result.get('error_type'):
            raise RuntimeError(str(async_task.result))
        else:
            return result

    def run(self, *args, **kwargs):

        try:
            res = self.func(*args, **kwargs)
        except ServiceError as e:
            self.controller.fail(e.code + 5000)
        except TaskError as e:
            raise InternalServerError(self.func_name, res.get_json_object())

        return res


class MissionFactory(dict):

    controller = None

    def __init__(self, params=None, controller=None):
        self.controller = controller
        if isinstance(params, dict):
            super().__init__(params)
        elif not params:
            super().__init__(dict())
        else:
            raise TypeError(
                f"Arguments data should be a 'dict' not {type(params)}.")

    def __getattr__(self, name):
        if name in self:
            attr = self.get(name)
            if isinstance(attr, dict):
                return self.__class__(attr, self.controller)
            else:
                return Mission(attr, self.controller)
        else:
            return self.__getattribute__(name)


class BaseController(RequestHandler):
    """Custom handler for other views module."""

    def __init__(self, application, request, **kwargs):
        super(BaseController, self).__init__(application, request, **kwargs)
        self.params = Arguments()
        self.token_args = Arguments()
        self.mg = MongoFactory()
        self.miss = MissionFactory(SERVICES, self)

    def _request_summary(self):
        s = ' '
        return f'{self.request.method.rjust(7, s)} {self.request.remote_ip.rjust(15, s)}  {self.request.path} '

    def log_exception(self, typ, value, tb):
        """Override to customize logging of uncaught exceptions.

        By default logs instances of `HTTPError` as warnings without
        stack traces (on the ``tornado.general`` logger), and all
        other exceptions as errors with stack traces (on the
        ``tornado.application`` logger).

        .. versionadded:: 3.1
        """
        if isinstance(value, HTTPError):
            if value.log_message:
                # format = "%d %s: " + value.log_message
                # args = ([value.status_code,
                #          self._request_summary()] + list(value.args))
                gen_log.warning('\033[0;31m' + value.log_message + '\033[0m')
        else:
            app_log.error(
                "Uncaught exception %s\n%r",
                self._request_summary(),
                self.request,
                exc_info=(typ, value, tb))

    async def options(self, *_args, **_kwargs):
        self.success()

    def prepare(self):
        params = dict(
            device=self.get_argument('device', 'web'),
            lang=self.get_argument('lang', 'cn').lower(),
            remote_ip=self.request.remote_ip,
            request_time=time.time())
        self.params = Arguments(params)

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Methods', 'GET,POST,PUT,DELETE')
        self.set_header('Access-Control-Allow-Headers',
                        'Authorization,Content-Type,Thor-Token')

    def get_token(self):
        header_name = O_O.server.token_header or 'Thor-Token'
        token = self.request.headers.get(header_name)
        try:
            return jwt.decode(token, O_O.server.token_secret)
        except jwt.DecodeError:
            return None

    def set_token(self, token_params: dict = None, expired: int = 3600):
        token_params = token_params or dict()
        token = jwt.encode(
            dict(token_params, timestamp=int(time.time()) + expired),
            O_O.server.token_secret)
        self.set_header(O_O.server.token_header or 'Thor-Token', token)

    def _check_params(self, args, checker):
        print(args, checker)
        for param_type in checker:
            # 设置默认值
            key = param_type.name
            if args.get(key) is None:
                if param_type.optional:
                    args[key] = param_type.default
                else:
                    raise MissingArgumentError(param_type.name)

            if param_type.type in (int, float, str, bool):
                if args[key] is None:
                    continue
                elif not isinstance(args[key], param_type.type):
                    try:
                        args[key] = param_type.type(args[key])
                    except:
                        raise ArgumentTypeError(key, param_type.type)
            elif param_type.type == list:
                if not isinstance(args[key], list):
                    raise ArgumentTypeError(key, list)
                for obj in args[key]:
                    self._check_params(obj, param_type.subparam)
            elif param_type.type == set:
                if not isinstance(args[key], dict):
                    raise ArgumentTypeError(key, dict)
                self._check_params(args[key], param_type.subparam)
            else:
                raise TypeError(f'Unkown type {type(checker[key])}')

    def check_params(self, checker=None):
        if self.request.method in ('GET', 'DELETE', 'HEAD'):
            args = self.parse_form_arguments(checker)
        else:
            args = self.parse_json_arguments(checker)

        self._check_params(args, checker)

        return args

    def parse_form_arguments(self, checker=None):
        """Parse FORM argument like `get_argument`."""
        if O_O.debug:
            dump_in(f'Input: {self.request.method} {self.request.path}',
                    self.request.body.decode()[:500])

        args = {
            k: v[0].decode() for k, v in self.request.arguments.items() if v[0]
        }

        return args

    def parse_json_arguments(self, checker=None):
        """Parse JSON argument like `get_argument`."""
        if O_O.debug:
            dump_in(f'Input: {self.request.method} {self.request.path}',
                    self.request.body.decode()[:500])

        try:
            args = json.loads(self.request.body.decode('utf-8'))
        except json.JSONDecodeError as exception:
            dump_error(self.request.body.decode())
            raise ParseJSONError(exception.doc)

        if not isinstance(args, dict):
            dump_error(self.request.body.decode())
            raise ParseJSONError('Request body should be a dictonary.')

        return args

    def finish_with_json(self, data):
        """Turn data to JSON format before finish."""
        self.set_header('Content-Type', 'application/json')

        if O_O.debug:
            if self.request.method == 'POST':
                info_list = [
                    f'\033[0;33mOutput: {self.request.method} {self.request.path}'
                ]
                if self.request.query:
                    query_list = [
                        f'\033[0;33m{i[0]:15s} {i[1]}'
                        for i in parse.parse_qsl(self.request.query)
                    ]
                    info_list.append('\n' + '\n'.join(query_list))
                if self.request.body:
                    try:
                        info_list.append('\n\033[0;33m' +
                                         self.request.body.decode())
                    except UnicodeDecodeError:
                        pass
                if data:
                    info_list.append('\n\033[0;33m' + json.dumps(data))
                dump_out(*info_list)

        raise Finish(json.dumps(data).encode())

    def fail(self, status, data=None, polyfill=None, **_kwargs):
        """assemble and return error data."""
        msg = get_status_message(status)
        self.finish_with_json(
            dict(status=status, msg=msg, data=data, **_kwargs))

    def success(self, data=None, msg='Successfully.', **_kwargs):
        """assemble and return error data."""
        self.finish_with_json(dict(status=0, msg=msg, data=data))

    async def fetch(self, api, method='GET', body=None, headers=None,
                    **_kwargs):
        """Fetch Info from backend."""
        body = body or dict()

        _headers = dict(host=self.request.host)
        if headers:
            _headers.update(headers)

        if '://' not in api:
            api = f'http://{O_O.server.back_ip}{api}'

        back_info = await httpclient.AsyncHTTPClient().fetch(
            api,
            method=method,
            headers=_headers,
            body=json.dumps(body),
            raise_error=False,
            allow_nonstandard_methods=True)

        res_body = back_info.body and back_info.body.decode() or None

        if back_info.code >= 400:
            return Arguments(
                dict(http_code=back_info.code, res_body=res_body, api=api))

        try:
            return Arguments(json.loads(res_body))
        except json.JSONDecodeError:
            raise ParseJSONError(res_body)



import re
import os
import types
import json
from unittest import TestCase

import requests
import yaml

from .utils import Arguments

from pprint import pprint


def pretty_print(prepared_request):
    """
    At this point it is completely built and ready
    to be fired; it is "prepared".

    However pay attention at the formatting used in 
    this function because it is programmed to be pretty 
    printed and may differ from the actual request.
    """
    method = prepared_request.method
    url = prepared_request.url
    headers = '\n'.join(
        f'{k}: {v}' for k, v in prepared_request.headers.items())
    body = prepared_request.body
    print('-----------START-----------')
    print(f'{method} {url}\n{headers}\n\n{body}')
    print('-----------ENDED-----------')


def get_tests():

    attrs = dict()

    attrs['checker'] = re.compile(r'(\$[a-z]+)( *: *)([^;]*)')
    attrs['number'] = re.compile(r'^[0-9]+$')
    attrs['word'] = re.compile(r'^\w+$')
    attrs['method'] = re.compile(r'(GET|POST|PUT|DELETE|HEAD|OPTIONS)')
    attrs['url'] = re.compile(r'(/[a-zA-Z0-9\-_]+)*/?')
    requests_data = read_config('test/requests.yml')
    env = Arguments(read_config('test/env.yml'))
    settings = Arguments(read_config('test/settings.yml'))

    if not isinstance(requests_data, list):
        raise ValueError('Wrong formatter.')

    attrs['requests_data'] = requests_data
    attrs['env'] = env
    attrs['settings'] = settings

    class_dict = {
        f'test_request_{i}': lambda self: TestItem(self, item, self.env).run()
        for i, item in enumerate(requests_data)
    }

    return type('TarpitTest', (TestCase,), dict(attrs, **class_dict))


def read_config(path):
    path = os.path.join(os.getcwd(), path)
    try:
        with open(path, 'r', encoding='utf-8') as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
    except FileNotFoundError:
        data = dict()

    return data


class TestItem():

    def __init__(self, tc, info, env):
        self.tc = tc
        self.info = info
        self.env = EnvRender(env)
        self.request = RequestBuilder()

    def run(self):
        info = self.env.render(self.info)
        info = Arguments(info)
        self.request.build(info.request)
        response = self.request.send()

        res_dict = dict(
            status_code=response.status_code,
            headers=dict(response.headers),
            body=response.json())

        checker = Checker(info.response, self.tc)
        checker.check(res_dict)

        if info.set_env:
            for k, v in info.set_env.items():
                self.set_env(k, v, res_dict)

        print(self.env)

    def set_env(self, name, value_path, resp):
        value_path_list = list(reversed(value_path.split('.')))
        value = self.find_value(value_path_list, resp)
        self.env.update(name, value)

    def find_value(self, target, data):
        if not target:
            return None
        cur = target.pop()
        if isinstance(data, dict):
            thisnode = data.get(cur)
        else:
            cur = int(cur)
            thisnode = data[cur]

        if not target:
            return thisnode

        if not isinstance(thisnode, (dict, list)):
            raise TypeError(
                f'Wrong type "{type(thisnode).__name__}" of node under key "{cur}"'
            )
        else:
            return self.find_value(target, thisnode)


class EnvRender():

    def __init__(self, env):
        self.env = env

    def update(self, k, v):
        self.env[k] = v

    def render(self, data):
        if isinstance(data, list):
            raise TypeError('item need to be a dict')
        return self._render_env_dict(data)

    def _render_env_dict(self, data):
        node = dict()
        for k, v in data.items():
            if isinstance(v, str):
                node[k] = v.format(**self.env)
            elif isinstance(v, dict):
                node[k] = self._render_env_dict(v)
            elif isinstance(v, list):
                node[k] = self._render_env_list(v)
            else:
                node[k] = v

        return node

    def _render_env_list(self, data):
        node = list()
        for item in data:
            if isinstance(item, str):
                node.append(item.format(**self.env))
            elif isinstance(item, dict):
                node.append(self._render_env_dict(item))
            elif isinstance(item, list):
                node.append(self._render_env_list(item))
            else:
                node.append(item)

        return node

    def __repr__(self):
        return json.dumps(self.env)


class RequestBuilder():

    def build(self, info):
        self.clear()
        self.headers = info.headers or Arguments()
        self.content_type = info.content_type or self.guess_content_type()

        data = dict()
        data['url'] = info.url
        data['method'] = info.method
        data['headers'] = dict(info.headers or dict())
        data['params'] = dict(info.query or dict())

        if isinstance(info.body, str):
            data['data'] = info.body
        elif self.content_type == 'json':
            data['json'] = info.body
        else:
            data['data'] = info.body

        self.prepared = requests.Request(**data).prepare()

    def send(self):
        if not self.prepared:
            raise ValueError('Request is not prepared.')
        return requests.Session().send(self.prepared)

    def clear(self):
        self.headers = None
        self.content_type = None
        self.prepared = None

    def guess_content_type(self):
        content_type = self.headers.get('Content-Type')

        if content_type == 'application/x-www-form-urlencoded':
            return 'urlencode'
        elif content_type == 'application/json':
            return 'json'
        else:
            return 'json'

    def raw_print(self):
        if not self.prepared:
            raise ValueError('Request is not prepared.')
        method = self.prepared.method
        url = self.prepared.url
        headers = '\n'.join(
            f'{k}: {v}' for k, v in self.prepared.headers.items())
        body = self.prepared.body
        print('-----------START-----------')
        print(f'{method} {url}\n{headers}\n\n{body}')
        print('-----------ENDED-----------')


class Checker():

    _expression = re.compile(r'(\$[a-z]+)( *: *)([^;]*)')

    def __init__(self, info, tc, start=None):
        self.tc = tc
        self._operator = {
            '$lt': lambda x, y: self.tc.assertLess(x, y),
            '$le': lambda x, y: self.tc.assertLessEqual(x, y),
            '$gt': lambda x, y: self.tc.assertGreater(x, y),
            '$ge': lambda x, y: self.tc.assertGreaterEqual(x, y),
            '$eq': lambda x, y: self.tc.assertEqual(x, y),
            '$ne': lambda x, y: self.tc.assertNotEqual(x, y),
            '$in': lambda x, y: self.tc.assertIn(x, y),
            '$is': lambda x, y: self.tc.assertIs(x, y),
            '$isnot': lambda x, y: self.tc.assertIsNot(x, y),
        }
        if not start:
            start = []
        elif isinstance(start, str):
            start = [start]

        self.info = info
        self.check_map = dict()
        self._parse(start, info)

    def _parse(self, cur, data):
        if isinstance(data, list):
            gen = enumerate(data)
        elif isinstance(data, dict):
            gen = data.items()
        else:
            raise TypeError(f'Unknown type "{type(data)}" of data.')

        for k, v in gen:
            path = cur + [str(k)]
            if isinstance(v, (str, int, float)):
                self.check_map['.'.join(path)] = v
            elif isinstance(v, (dict, list)):
                self._parse(path, v)
            else:
                raise TypeError(f'Unknown type "{type(v)}" of key "{k}"')

    def check(self, resp):
        for k, checker in self.check_map.items():
            value = self.get_value(k, resp)
            self.check_item(checker, value)

    def get_value(self, path, resp):
        value = resp
        for p in path.split('.'):
            if p in value:
                value = value[p]
            elif p.isdigit():
                i = int(p)
                if not isinstance(value, list):
                    return 1
                elif len(value) <= i:
                    return 2
                else:
                    value = value[p]
            else:
                return None
        return value

    def check_item(self, exp, checkee):
        matcher = self._expression.match(exp)
        if not matcher:
            raise ValueError(f'Unkown expression {exp}')

        operator, _, answer = matcher.groups()
        answer = eval(answer, {'__buildins__': None})

        return self.test(operator, checkee, answer)

    def test(self, operator, a, b):
        if operator not in self._operator:
            raise KeyError(f'Unknown operator "{operator}"')
        return self._operator[operator](a, b)



# coding:utf-8

from .logger import dump_in, dump_out, dump_error
from .arguments import Arguments
from .totp import get_totp_secret, get_totp_token
from .json_encoder import tarpit_dumps, tarpit_loads
from .regex import Patterns
from .passwd import encode_passwd, generate_id

__all__ = [
    'dump_in',
    'dump_out',
    'dump_error',
    'Arguments',
    'get_totp_secret',
    'get_totp_token',
    'tarpit_dumps',
    'tarpit_loads',
    'generate_id',
    'encode_passwd',
]


# coding:utf-8
"""Library Module."""
import os
import argparse
import unittest

from .command import serve, generate
from .test import get_tests

def main():

    parser = argparse.ArgumentParser(description='Tarpit commend line tools.')
    subparsers = parser.add_subparsers(help='sub-command help', dest='command')

    subparsers.add_parser('serve')

    subparsers.add_parser('test')

    _worker_parser = subparsers.add_parser('worker')

    # cc = CeleryConfig()
    # worker_parser.add_argument(
    #     '-q',
    #     '--queue',
    #     type=str,
    #     choices=[*cc.queue_list],
    #     help='Run worker instance.')

    create_parser = subparsers.add_parser('create')
    create_parser.add_argument('directory', type=str, help='target directory.')

    params = parser.parse_args()

    if params.command == 'serve':
        serve()
    elif params.command == 'worker':
        print('runworker')
    elif params.command == 'create':
        if os.path.exists(params.directory):
            parser.error('directory exists')
        generate(params.directory)
    elif params.command == 'test':
        suite = unittest.TestLoader().loadTestsFromTestCase(get_tests())
        unittest.TextTestRunner().run(suite)
    else:
        parser.error('No valid options.')